/*
 * Copyright (c) 2001-2002 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the XLED software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * XLED source distribution.
 *
 * $Id: xled_view.c,v 1.2 2002/12/11 19:06:38 brbarret Exp $
 * 
 *	Function:	- overall application view in main window
 */

#define _NO_PROTO

#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/ScrolledW.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <events.h>
#include <ledreq.h>
#include <net.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <xled.h>

#include "lamlogo.xpm"

/*
 * global functions
 */
Widget			xled_view_build();
void			xled_view_clear();
void			xled_view_init();
void			xled_view_onoff();
void			xled_view_state();

/*
 * external functions
 */
extern void		xled_seticon();
extern void		xled_setminmax();
extern Pixmap		xpm_build();

/*
 * external variables
 */
extern AppData		app_res;		/* application resources */

/*
 * local functions
 */
static int		fn_width();
static int		fn_height();
static void		cb_draw();
static void		drawclear();
static void		gccreate();
static void		initledreg();
static void		plotproc();
static void		viewfill();
static void		viewplot();

/*
 * local variables
 */
static struct xlproc	*xledprocs = 0;	/* array of processes */
static int		xlednprocs = 0;	/* # of processes */
static int		xlednxprocs;	/* # procs horizontally */
static int		xlednyprocs;	/* # procs vertically */
static int		drawwidth;	/* drawing area width */
static int		drawheight;	/* drawing area heigth */
static int		fl_autoclr = 1;	/* auto-clear-screen flag */
static int		xledregx[LEMAXNLEDS][XLED_NPOINTS];
static int		xledregy[LEMAXNLEDS][XLED_NPOINTS];
static unsigned int	depth;		/* screen depth */
static Widget		draw_w;		/* drawing area widget */
static Display		*disp;		/* drawing area display */
static Drawable		root;		/* root window */
static Pixmap		pixmap;		/* copy of drawing area */
static GC		fggc;		/* foreground pen */
static GC		bggc;		/* background pen */
static GC		boldgc;		/* bold pen */
static GC		xledgc[LEMAXNLEDS];

/*
 *	xled_view_build
 *
 *	Function:	- build widgets for graphics viewing subwindow
 *	Accepts:	- parent widget
 *	Returns:	- created widget
 */
Widget
xled_view_build(parent_w)

Widget			parent_w;

{
	Widget		view_w;		/* subwindow widget */
	Widget		scroll_w;	/* scrolled window widget */

	view_w = XtVaCreateWidget("view_mgr",
			xmFormWidgetClass, parent_w,
			NULL);
/*
 * Build the scrolled window.
 */
	xlednxprocs = XLED_DEFXNPROCS;
	xlednyprocs = XLED_DEFYNPROCS;
	drawwidth = fn_width(xlednxprocs);
	drawheight = fn_height(xlednyprocs);

	scroll_w = XtVaCreateManagedWidget("view_scrwin",
			xmScrolledWindowWidgetClass, view_w,
			XmNwidth, drawwidth + XLED_DRAWXEXTRA,
			XmNheight, drawheight + XLED_DRAWYEXTRA,
			XmNscrollingPolicy, XmAUTOMATIC,
			XmNspacing, 0,
			XmNtopAttachment, XmATTACH_FORM,
			XmNtopOffset, 2,
			XmNbottomAttachment, XmATTACH_FORM,
			XmNleftAttachment, XmATTACH_FORM,
			XmNrightAttachment, XmATTACH_FORM,
			NULL);

	xled_setminmax(fn_width(1), fn_height(1), drawwidth, drawheight);
/*
 * Build the drawing area.
 */
	draw_w = XtVaCreateManagedWidget("view_draw",
			xmDrawingAreaWidgetClass, scroll_w,
			XmNwidth, drawwidth,
			XmNheight, drawheight,
			NULL);
	XtAddCallback(draw_w, XmNexposeCallback, cb_draw, NULL);

	XtManageChild(draw_w);

	XtManageChild(scroll_w);

	XtManageChild(view_w);
/*
 * Create the graphic contexts.
 */
	gccreate();
/*
 * Initialize LED regions.
 */
	initledreg();

	return(view_w);
}

/*
 *	gccreate
 *
 *	Function:	- create graphic contexts
 */
static void
gccreate()

{
	Pixel		fg, bg;		/* colours of drawing area */
	XGCValues	gcval;		/* graphic context info */
	Pixmap		logopix;	/* logo pixmap */
	XColor		led_def;	/* LED colour definition */
	XColor		white_def;	/* white colour definition */
	XColor		dev_null;	/* unused functionality */
	int		logowidth;	/* width of LAM logo */
	int		logoheight;	/* height of LAM logo */
	int		i;		/* favourite index */

	disp = XtDisplay(draw_w);
	root = RootWindowOfScreen(XtScreen(draw_w));
	depth = DefaultDepthOfScreen(XtScreen(draw_w));
/*
 * Create the foreground/background graphic contexts.
 */
	XtVaGetValues(draw_w, XmNforeground, &fg, XmNbackground, &bg, NULL);

	gcval.foreground = fg;
	gcval.background = bg;

	fggc = XCreateGC(disp, root, (GCForeground | GCBackground), &gcval);

	gcval.foreground = bg;
	bggc = XCreateGC(disp, root, GCForeground, &gcval);

	XAllocNamedColor(disp, DefaultColormapOfScreen(XtScreen(draw_w)),
			"white", &white_def, &dev_null);

	gcval.foreground = white_def.pixel;
	gcval.background = bg;
	gcval.font = app_res.ap_rankfont->fid;
	boldgc = XCreateGC(disp, root,
			(GCFont | GCForeground | GCBackground), &gcval);
/*
 * Create the LEDs graphic contexts.
 */
	for (i = 0; i < LEMAXNLEDS; ++i) {
		XAllocNamedColor(disp,
			DefaultColormapOfScreen(XtScreen(draw_w)),
			app_res.ap_ledcolour[i], &led_def, &dev_null);

		gcval.foreground = led_def.pixel;
		gcval.background = bg;
		xledgc[i] = XCreateGC(disp, root,
				(GCForeground | GCBackground), &gcval);
	}
/*
 * Create the pixmap where the drawing area is saved.
 */
	pixmap = XCreatePixmap(disp, root, drawwidth, drawheight, depth);

	drawclear();
/*
 * Create the LAM logo and set it as the icon.
 */
	xpm_build(draw_w, lamlogo_xpm, (Pixel) 0, &logopix, &logowidth,
		&logoheight);

	if (logopix != XmUNSPECIFIED_PIXMAP) {
		xled_seticon(logopix);
/*
 * Copy the logo to the initial screen.
 */
		if ((logowidth <= drawwidth) &&
				(logoheight <= drawheight)) {
			XCopyArea(disp, logopix, pixmap, fggc,
				0, 0, logowidth, logoheight,
				(drawwidth - logowidth) / 2,
				(drawheight - logoheight) / 2);
		}
	}
}

/*
 *	cb_draw
 *
 *	Function:	- callback function for drawing area events
 *			- handles expose events
 *	Accepts:	- widget
 *			- client data
 *			- ptr callback struct
 */
static void
cb_draw(w, cdata, p_callback)

Widget			w;
XtPointer		cdata;
XmDrawingAreaCallbackStruct
			*p_callback;

{
	XEvent		*event;		/* activating event */

	event = p_callback->event;
/*
 * Handle the expose event.
 */
	if (p_callback->reason == XmCR_EXPOSE) {
		XCopyArea(event->xany.display, pixmap, event->xany.window,
				fggc, 0, 0, drawwidth, drawheight, 0, 0);
	}
}

/*
 *	drawclear
 *
 *	Function:	- clear the drawing area
 */
static void
drawclear()

{
	XFillRectangle(disp, pixmap, bggc, 0, 0, drawwidth, drawheight);
}

/*
 *	xled_view_clear
 *
 *	Function:	- clear the view window
 */
void
xled_view_clear()

{
	drawclear();
	viewplot();
	XCopyArea(disp, pixmap, XtWindow(draw_w), fggc,
				0, 0, drawwidth, drawheight, 0, 0);
	fl_autoclr = 0;
}

/*
 *	xled_view_init
 *
 *	Function:	- initalize LEDs
 *	Accepts:	- process rank
 *			- # processes
 */
void
xled_view_init(proc, nprocs)

int4			proc;
int4			nprocs;

{
/*
 * Clear the screen the first time around.
 */
	if (fl_autoclr) xled_view_clear();
/*
 * Set/expand/shrink the # processes.
 */
	viewfill(nprocs);
/*
 * Update the process LEDs to be all off.
 */
	xled_view_onoff(proc, nprocs, (int4) -1, 0);
}

/*
 *	xled_view_onoff
 *
 *	Function:	- turn LEDs on or off
 *	Accepts:	- process rank
 *			- # processes
 *			- LED mask
 *			- on flag
 */
void
xled_view_onoff(proc, nprocs, ledmask, fl_on)

int4			proc;
int4			nprocs;
int4			ledmask;
int			fl_on;

{
	int		i, j;		/* favourite indices */
	int		bitmask;	/* single bit mask */
	int		minx, miny;	/* min coordinates */
	int		maxx, maxy;	/* max coordinates */
	int		fl_mod;		/* pixmap modified flag */
	struct xlproc	*p;		/* favourite pointer */
	XPoint		*pp;		/* another pointer */
	XPoint		points[XLED_NPOINTS];	/* region points */
/*
 * Update the process LEDs.
 */
	if ((proc < 0) || (proc >= xlednprocs)) return;

	p = xledprocs + proc;

	if (fl_on) {
		p->xlp_ledmask |= ledmask;
	} else {
		p->xlp_ledmask &= ~ledmask;
	}
/*
 * Plot the LED regions.
 */
	minx = drawwidth + 1;
	miny = drawheight + 1;
	maxx = maxy = -1;
	fl_mod = 0;

	for (i = 0, bitmask = 1; i < LEMAXNLEDS; ++i, bitmask <<= 1) {

		if (ledmask & bitmask) {
			fl_mod = 1;

			pp = points;
			for (j = 0; j < XLED_NPOINTS; ++j, ++pp) {

				pp->x = p->xlp_xcenter + xledregx[i][j];
				pp->y = p->xlp_ycenter + xledregy[i][j];

				if (minx > pp->x) minx = pp->x;
				if (miny > pp->y) miny = pp->y;
				if (maxx < pp->x) maxx = pp->x;
				if (maxy < pp->y) maxy = pp->y;
			}

			XFillPolygon(disp, pixmap,
					fl_on ? xledgc[i] : bggc,
					points, XLED_NPOINTS,
					Convex, CoordModeOrigin);
		}
	}

	if (fl_mod) {
		XCopyArea(disp, pixmap, XtWindow(draw_w), fggc,
				minx, miny, maxx - minx + 1,
				maxy - miny + 1, minx, miny);
	}
}

/*
 *	xled_view_state
 *
 *	Function:	- get LED state information
 *	Accepts:	- proces rank
 *			- ptr # LEDs (returned value)
 *			- ptr LEDs mask (returned value)
 */
void
xled_view_state(proc, pnleds, pledmask)

int4			proc;
int4			*pnleds;
int4			*pledmask;

{
	struct xlproc	*p;		/* favourite pointer */

	if ((proc < 0) || (proc >= xlednprocs)) {
		*pnleds = *pledmask = 0;
	}
	else {
		p = xledprocs + proc;
		*pnleds = LEMAXNLEDS;
		*pledmask = p->xlp_ledmask;
	}
}

/*
 *	viewfill
 *
 *	Function:	- fill the processes in the viewing subwindow
 *			- create the processes data structure
 *	Accepts:	- # of processes
 */
static void
viewfill(nprocs)

int			nprocs;

{
	int		x, y;		/* pixel coordinates */
	int		startx;		/* starting x ordinate */
	int		side;		/* sqrt(3)/2 * radius */
	int		delta;		/* distance between procs */
	int		i, j;		/* favourite indices */
	Dimension	swidth;		/* scrolled window width */
	Dimension	sheight;	/* scrolled window height */
	struct xlproc	*p;		/* favourite pointer */

	if ((nprocs == 0) || (nprocs == xlednprocs)) {
		return;
	}
/*
 * Allocate the process data structure.
 */
	if (xledprocs) {
		free((char *) xledprocs);
	}

	xledprocs = (struct xlproc *)
			malloc((unsigned) nprocs * sizeof(struct xlproc));
	if (xledprocs == 0) {
		terror("xled (malloc)");
		kexit(errno);
	}

	xlednprocs = nprocs;
/*
 * Layout the processes in a "squarish" area.
 */
	for (i = 0, j = 0; j < nprocs; ++i) {
		j += i + i + 1;
	}

	xlednxprocs = i;
	xlednyprocs = (nprocs + xlednxprocs - 1) / xlednxprocs;
/*
 * Compute size of new drawing area.
 */
	side = (SQRT3 * XLED_PROCRADIUS / 2) + 0.5;

	drawwidth = fn_width((xlednxprocs < XLED_DEFXNPROCS) ?
					XLED_DEFXNPROCS : xlednxprocs);
	drawheight = fn_height((xlednyprocs < XLED_DEFYNPROCS) ?
					XLED_DEFYNPROCS : xlednyprocs);
/*
 * Shrink the scrolled window if needed.
 */
	XtVaGetValues(XtParent(draw_w),
			XmNwidth, &swidth, XmNheight, &sheight, NULL);

	if ((swidth > (drawwidth + XLED_DRAWXEXTRA)) ||
			(sheight > (drawheight + XLED_DRAWYEXTRA))) {
		xled_setsize(drawwidth, drawheight);
	}
/*
 * Restrict the min/max top window sizes.
 */
	xled_setminmax(fn_width(1), fn_height(1), drawwidth, drawheight);
/*
 * Resize the drawing area and its pixmap.
 */
	XtVaSetValues(draw_w,
		XmNwidth, drawwidth, XmNheight, drawheight, NULL);

	XFreePixmap(disp, pixmap);

	pixmap = XCreatePixmap(disp, root, drawwidth, drawheight, depth);
/*
 * Fill the processes data structure.
 */
	startx = XLED_PROCRADIUS + XLED_XMARGIN;

	if (xlednxprocs < XLED_DEFXNPROCS) {
		startx += (drawwidth - fn_width(xlednxprocs)) / 2;
	}

	x = startx;

	y = side + XLED_YMARGIN;

	if (xlednyprocs < XLED_DEFYNPROCS) {
		y += (drawheight - fn_height(xlednyprocs)) / 2;
	}

	delta = 3 * XLED_PROCRADIUS / 2;

	p = xledprocs;

	for (i = j = 0; i < nprocs; ++i, ++p) {

		p->xlp_xcenter = x;
		p->xlp_ycenter = y + ((j & 1) ? side : 0);
		p->xlp_ledmask = 0;

		if (++j == xlednxprocs) {
			x = startx;
			y += side + side;
			j = 0;
		} else {
			x += delta;
		}
	}

	drawclear();
	viewplot();
}

/*
 *	fn_width
 *
 *	Function:	- compute pixel-width of drawing area
 *	Accepts:	- # processes horizontally
 *	Returns:	- width
 */
static int
fn_width(nprocs)

int			nprocs;

{
	int		width;

	width = (2 * XLED_XMARGIN) +
			(((3 * nprocs) + 1) * XLED_PROCRADIUS / 2);
	return(width);
}

/*
 *	fn_height
 *
 *	Function:	- compute pixel-height of drawing area
 *	Accepts:	- # processes vertically
 *	Returns:	- height
 */
static int
fn_height(nprocs)

int			nprocs;

{
	int		side;
	int		height;

	side = (SQRT3 * XLED_PROCRADIUS / 2) + 0.5;
	height = (((2 * nprocs) + 1) * side) + (2 * XLED_YMARGIN);
	return(height);
}

/*
 *	viewplot
 *
 *	Function:	- plot all processes
 */
static void
viewplot()

{
	int		i;

	for (i = 0; i < xlednprocs; ++i) {
		plotproc(i);
	}

	XCopyArea(disp, pixmap, XtWindow(draw_w), fggc,
				0, 0, drawwidth, drawheight, 0, 0);
}

/*
 *	plotproc
 *
 *	Function:	- plot the process hexagon
 *	Accepts:	- process rank
 */
static void
plotproc(rank)

int			rank;

{
	int		x, y;
	int		side;
	int		strwidth;
	XPoint		points[7];
	char		buf[20];
/*
 * Draw the process hexagon.
 */
	x = xledprocs[rank].xlp_xcenter;
	y = xledprocs[rank].xlp_ycenter;

	side = (SQRT3 * XLED_PROCRADIUS / 2) + 0.5;

	points[0].x = x - XLED_PROCRADIUS; points[0].y = y;
	points[1].x = x - (XLED_PROCRADIUS/2); points[1].y = y - side;
	points[2].x = x + (XLED_PROCRADIUS/2); points[2].y = y - side;
	points[3].x = x + XLED_PROCRADIUS; points[3].y = y;
	points[4].x = x + (XLED_PROCRADIUS/2); points[4].y = y + side;
	points[5].x = x - (XLED_PROCRADIUS/2); points[5].y = y + side;
	points[6].x = x - XLED_PROCRADIUS; points[6].y = y;

	XDrawLines(disp, pixmap, fggc, points, 7, CoordModeOrigin);
/*
 * Write the process rank.
 */
	sprintf(buf, "%d", rank);

	strwidth = XTextWidth(app_res.ap_rankfont, buf, strlen(buf));

	XDrawString(disp, pixmap, boldgc,
			x - (strwidth / 2),
			y + (app_res.ap_rankfont->ascent) / 2,
			buf, strlen(buf));
}

/*
 *	initledreg
 *
 *	Function:	- initialize the LED regions
 *			- sets coordinates of 6 LED triangles in a hexagon
 */
static void
initledreg()
 
{
/*
 * Set LED #0 triangle.
 */
	xledregx[0][0] = 0;
	xledregy[0][0] = XLED_CLEARANCE * XLED_PROCRADIUS;

	xledregx[0][1] = (1 - (XLED_CLEARANCE * SQRT3)) * XLED_PROCRADIUS / 2;
	xledregy[0][1] = (SQRT3 - XLED_CLEARANCE) * XLED_PROCRADIUS / 2;

	xledregx[0][2] = - xledregx[0][1];
	xledregy[0][2] = xledregy[0][1];
/*
 * Set LED #1 triangle.
 */
	xledregx[1][0] = - (SQRT3 * XLED_CLEARANCE * XLED_PROCRADIUS / 2);
	xledregy[1][0] = XLED_CLEARANCE * XLED_PROCRADIUS / 2;

	xledregx[1][1] = - (XLED_PROCRADIUS / 2);
	xledregy[1][1] = (SQRT3 - (2 * XLED_CLEARANCE)) * XLED_PROCRADIUS / 2;

	xledregx[1][2] = ((XLED_CLEARANCE * SQRT3) - 2) * XLED_PROCRADIUS / 2;
	xledregy[1][2] = xledregy[1][0];
/*
 * Set LED #2 triangle.
 */
	xledregx[2][0] = xledregx[1][0];
	xledregy[2][0] = - xledregy[1][0];

	xledregx[2][1] = xledregx[1][2];
	xledregy[2][1] = xledregy[2][0];

	xledregx[2][2] = xledregx[1][1];
	xledregy[2][2] = - xledregy[1][1];
/*
 * Set LED #3 triangle.
 */
	xledregx[3][0] = 0;
	xledregy[3][0] = - xledregy[0][0];

	xledregx[3][1] = - xledregx[0][1];
	xledregy[3][1] = - xledregy[0][1];

	xledregx[3][2] =  xledregx[0][1];
	xledregy[3][2] = xledregy[3][1];
/*
 * Set LED #4 triangle.
 */
	xledregx[4][0] = - xledregx[1][0];
	xledregy[4][0] = xledregy[2][0];

	xledregx[4][1] = - xledregx[1][1];
	xledregy[4][1] = xledregy[2][2];

	xledregx[4][2] = - xledregx[1][2];
	xledregy[4][2] = xledregy[4][0];
/*
 * Set LED #5 triangle.
 */
	xledregx[5][0] = xledregx[4][0];
	xledregy[5][0] = xledregy[1][0];

	xledregx[5][1] = xledregx[4][2];
	xledregy[5][1] = xledregy[1][0];

	xledregx[5][2] = xledregx[4][1];
	xledregy[5][2] = xledregy[1][1];
}
